#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""A demo for running FiGTNE on a randomly generated data."""

import sys
import warnings
import numpy as np
import random_walks
from tqdm import tqdm
import networkx as nx
from sklearn import preprocessing
from gensim.models import Word2Vec
from sklearn.metrics import f1_score
from multiprocessing import cpu_count
from sklearn.linear_model import LogisticRegression
from sklearn.model_selection import train_test_split

warnings.filterwarnings('ignore')

def create_temporal_graph_for_node_classification(dataset_filename, is_directed=True):
    """
    Create a graph and labels for all nodes

    @param dataset_filename: str, the filename of the edgelist

    @return a nx.MultiGraph (edge: [source, end, {'time': timestampe, 'weight':1}])
            a numpy array (shape: Nx2, each row is: node_id, node_label)
    """

    print('Reading edgelist...')
    
    if 'demo' in dataset_filename:
        ## the demo.npy is with four columns: source_id, target_id, label of target_id, timestamp (each row denotes an edge)
        
        data = np.load(dataset_filename)
    
        source_id = data[:,0]
        target_id = data[:,1]
        time_stamp = data[:,-1]
        source_id += np.max(target_id)
    
        #target_id and corresponding label
        node_label_dict = {}
        for idx in tqdm(range(source_id.shape[0])):
            if data[idx,1] not in node_label_dict.keys():
                node_label_dict[data[idx,1]] = [data[idx,1],data[idx,2]]
        node_label = np.array(list(node_label_dict.values()))
        print('The number of nodes is %d'%(node_label.shape[0]))
    
        node_set = set()
        edgelist_time = [] #each element is "source,end,timestamp"
    
        for idx in tqdm(range(source_id.shape[0])):
            source = int(source_id[idx])
            end = int(target_id[idx])
    
            node_set.add(source)
            node_set.add(end)
    
            #initial the weight of each edge to 1
            edgelist_time.append([source, end, {'time':int(time_stamp[idx]), 'weight':1}])
    
        print('>>> Completed!\n')
    
        #create the graph for embedding
        print('Creating graph...')
        graph = nx.MultiGraph()
        graph.add_edges_from(edgelist_time) 
        print('>>> Completed!\n')
    
        return graph, node_label      
    else:
        print('The import of the input dataset has not been supported!')
        sys.exit()

def learn_embeddings(walks, args):
    '''
    Learn embeddings by optimizing the Skipgram objective using SGD.

    @param walks: a list, includes [num_nodes*num_walk_each_node]*walk_length walks

    @return a word_vec dict defined in GENSIM
    '''
    walks = [list(map(str, walk)) for walk in walks]
    model = Word2Vec(walks, size=args['dimensions'], window=args['window_size'], 
                     min_count=0, sg=1, workers=args['workers'], iter=args['epoch'])

    return model.wv.word_vec

def node_classification(word_vec, id_and_label, train_size=0.5):
    """
    Test the embedding representations on the task of node classification
    
    @param word_vec: a dict that saves embeddings of all nodes
    @param id_and_label: numpy array (Nx2), each row is id and label
    @param train_size: train_data size / all_data size
    
    @return a dict that saves the results of evaluation
    """
    
    ##Get the vectors of targets
    node_vec = []
    for i in range(id_and_label.shape[0]):
        node_vec.append(word_vec(str(id_and_label[i,0])))
    node_vec = np.stack(node_vec, axis=0)   
    node_vec = preprocessing.scale(node_vec)

    node_label = id_and_label[:,1]    
    clf = LogisticRegression()
    train_vec, test_vec, train_label, test_label = train_test_split(node_vec, node_label, train_size=train_size,
                                                                    test_size=(1-train_size), random_state=1)
    
    model = clf.fit(train_vec, train_label)

    prediction = model.predict(test_vec)
    f1_macro = f1_score(y_true=test_label, y_pred=prediction, average='macro')
    f1_micro = f1_score(y_true=test_label, y_pred=prediction, average='micro')
    
    return {'f1_macro': f1_macro, 'f1_micro': f1_micro}  

def TNE(args):
    ##create a temporal graph
    if args['task'] == 'Node':
        graph, node_label = create_temporal_graph_for_node_classification(
            dataset_filename=args['edgelist_file'], is_directed=args['directed'])

    ##generate random walks
    walks   = [] #walks without timestamps, which are the input of GENSIM (node2vec)
    walks_T = [] #walks with timestamps, which are the input for modifying transition matrix
    G = random_walks.TemporalGraph(graph, args['context'])
    G.preprocess_transition_probs() # the weight of each edge is initialized to 1
    for i in range(args['num_walks']):
        print('%d-th walk...' %(i+1))
        walks_for_embedding, walks_with_timestamp = G.simulate_walks(1, args['walk_length'])
        if args['change_prob']:
            if np.mod(i, args['K']) == 0:
                G.change_prob(walks_with_timestamp, args['context'], args['delta'])

        walks.extend(walks_for_embedding)
        walks_T.extend(walks_with_timestamp)

    ##generate node embeddings based on random walks
    word_vec = learn_embeddings(walks, args)

    ##evaluation
    result_dict = node_classification(word_vec, node_label, args['train_size'])

    return result_dict

if __name__ == '__main__':
    
    def create_args():
        args = {}
        """
        Current options:
           './data/demo.npy'
        """
        args['edgelist_file']  = './data/demo.npy'
        
        ##node2vec related hyperparameters
        args['dimensions']     = 64
        args['walk_length']    = 20
        args['num_walks']      = 10
        args['window_size']    = 3
        args['epoch']          = 1
        
        ##general p1arameters
        args['workers']        = cpu_count()
        args['weighted']       = True
        args['directed']       = False
        
        ##temporal network embedding related parameters
        #which temporal context is used: 'closeness'
        args['context']        = 'closeness'
        
        #whether generate reinforced random walks
        #only initial transition probability matrix is used if this is set to False
        args['change_prob']    = True        
        
        #which task is tested using embedding representations
        #'Node': node classification
        args['task']           = 'Node'
        
        #size of training data
        args['train_size']     = 0.6
        
        #interval of changing edge weight
        args['K']              = 1
        
        #the amount of each edge weight increase
        args['delta']          = 1

        return args

    args = create_args()
    result_dict = TNE(args)
    
    print('>>> Performance:')
    for k, v in result_dict.items():
        print('\t%s: %.5f' %(k, v))