#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import sys  
import random
import numpy as np
import networkx as nx
from tqdm import tqdm
from scipy.stats import rankdata
from multiprocessing import cpu_count, Pool, Process, Manager

class TemporalGraph():
    def __init__(self, nx_G, context='closeness'):
        self.G = nx_G
        self.is_directed = nx_G.is_directed()
        self.context = context
        self.calculate_temporal_transition_matrix(context)

    def node2vec_walk(self, walk_length, start_node):
        '''
        Simulate a random walk starting from start node.
        '''
        G = self.G
        walk = [start_node]
        
        walk_with_timestamp = [{'index_of_in_node': None,
                                'index_of_out_node': None,
                                'cur_node': start_node,
                                'out_node': start_node,
                                'time': 0}] 

        while len(walk) < walk_length:
            cur = walk[-1]
            cur_prob_matrix = self.transition_prob_dict[cur][0]
            if cur_prob_matrix.shape[1] > 0:
                in_node_index = self.transition_prob_dict[cur][1]
                out_node = self.transition_prob_dict[cur][2]   
                out_node_timestamp = self.transition_prob_dict[cur][3]
                
                if len(walk) == 1:
                    #now assume the timestamp of first node is 0, and randomly choose a out_node
                    prev_idx = np.asscalar(np.random.choice(cur_prob_matrix.shape[0], 1))
                else:
                    prev_idx = in_node_index[walk[-2]]
                    
                sampling_prob = cur_prob_matrix[prev_idx,:]
                if np.sum(sampling_prob) > 0:
                    sampling_prob = sampling_prob / np.sum(sampling_prob)
                    tmp_index = np.asscalar(np.random.choice(np.arange(len(out_node)),
                                                             1, p=sampling_prob.tolist()))
                    walk.append(out_node[tmp_index])
                    walk_with_timestamp.append({'index_of_in_node': prev_idx,
                                                'index_of_out_node': tmp_index, 
                                                'cur_node': cur,
                                                'out_node': out_node[tmp_index],
                                                'time': out_node_timestamp[tmp_index]})
                else:
                    break
            else:
                break

        return walk, walk_with_timestamp

    def simulate_walks(self, num_walks, walk_length):
        '''
        Repeatedly simulate random walks from each node.
        '''
        G = self.G
        walks = []
        walks_with_timestamp = []
        nodes = list(G.nodes())
        
        def sub_process(process_idx, walk_length, node_list, node_queue):
            result = [[],[]]
            for node in tqdm(node_list):
                tmp = self.node2vec_walk(walk_length=walk_length, start_node=node)
                result[0].append(tmp[0])
                result[1].append(tmp[1])               
            node_queue.put(result) 
        
        def chunks(l, n):
            chunk_size = int(len(l)/n)+1
            for i in range(0, len(l), chunk_size):
                yield l[i:i+chunk_size]        
        
        process_num = int(cpu_count())
        
        for walk_iter in range(num_walks):
            node_queue = Manager().Queue()
            
            random.shuffle(nodes)
            nodes_chunk = list(chunks(nodes, process_num))
            
            jobs = []
            for i in range(process_num):
                p = Process(target=sub_process, args=(i,walk_length,nodes_chunk[i],node_queue))
                p.start()
                jobs.append(p)
            
            for i in range(process_num):
                jobs[i].join()
            
            tmp_result = []  
            while len(tmp_result) != process_num:
                tmp_result.append(node_queue.get())          
            
            for tmp in tmp_result:
                walks.extend(tmp[0])
                walks_with_timestamp.extend(tmp[1])
        
        return walks, walks_with_timestamp

    def preprocess_transition_probs(self):
        '''
        Preprocessing of transition probabilities for guiding the random walks.
        '''
        G = self.G
        is_directed = self.is_directed

        self.transition_prob_dict = {}
        print('Preprocess transition probs...')
        for node in tqdm(G.nodes()):  
            if not is_directed:
                degree = len(G.edges(node)) # G.degree(node), that would be a problem if a self-loop appears (one edge but adding two degress)
                    
                prob = np.zeros((degree, degree))
                in_node_index = {} # a dict for mapping (e.g. 20:1, node 20 is the nbr[1] of current node)
                out_node = [] # out node of current node
                out_node_timestamp = [] #timestamps of [edge(current_node,x) for x in out_node]
                
                for idx,tmp_in_edge in enumerate(G.edges(node)):
                    in_node_index[tmp_in_edge[1]] = idx
                for idx,tmp_out_edge in enumerate(G.edges(node, data=True)):
                    out_node.append(tmp_out_edge[1])
                    prob[:,idx] = tmp_out_edge[2]['weight'] # initialize to the weight of the edge
                    out_node_timestamp.append(tmp_out_edge[2]['time'])
                
                self.transition_prob_dict[node] = [prob, in_node_index, out_node, out_node_timestamp]
        print('>>> Completed!\n')

    def change_prob(self, walks, context='closeness', delta=1):
        """
        Here is for modifying transition prob matrix (TRRW) based on some pre-defined rules
        """        
        for one_walk in walks:
            for i in range(1,len(one_walk)-1):
                    
                # both indices starts from 0 to (in_degree/out_degree)-1
                pre_node_index = one_walk[i]['index_of_in_node']
                out_node_index = one_walk[i]['index_of_out_node']
                current_node = one_walk[i]['cur_node']
                
                tmp = self.transition_prob_dict[current_node][0]
                
                # if current prob does not equal to the target prob, add 'delta' on the weight of transition
                if (tmp[pre_node_index][out_node_index] + delta) < \
                   self.P_t[current_node][pre_node_index][out_node_index]:
                    tmp[pre_node_index][out_node_index] += delta
                else:
                    tmp[pre_node_index][out_node_index] = self.P_t[current_node][pre_node_index][out_node_index]
                    
                self.transition_prob_dict[current_node][0] = tmp

    def calculate_temporal_transition_matrix(self, context='closeness'):
        """
        Given one temporal context, calculate the corresponding temporal transition matrix. Each node
        would have a transition matrix with the shape of in_degree*out_degree.
        
        Contexts are supported:
        1. Closeness: The weight of each edge is equal to the rank (in increasing order) of abs(t_out - t_in);
        2. Unbiased: The weight of each edge is 1.
        """
        
        self.P_t = {}
        is_directed = self.is_directed
        G = self.G
        
        print('Precalculate temporal transition probs...')
        for node in tqdm(G.nodes()):   
            
            in_node_index  = {} # a dict for mapping (e.g. 20:1, node 20 is the nbr[1] of current node)
            out_node_index = {} # same with above
            in_node_timestamp  = [] # record the timestamps of in_nodes
            out_node_timestamp = [] # same with above                   
            
            # get the timestamps and node_index of in_edge and out_edge
            if is_directed:                                    
                for idx,tmp_in_edge  in enumerate(G.in_edges(node, data=True)):
                    in_node_index[tmp_in_edge[0]] = idx
                    in_node_timestamp.append(tmp_in_edge[2]['time'])
                for idx,tmp_out_edge in enumerate(G.out_edges(node, data=True)):
                    out_node_index[tmp_out_edge[1]] = idx
                    out_node_timestamp.append(tmp_out_edge[2]['time'])  
            else:
                for idx,tmp_in_edge  in enumerate(G.edges(node, data=True)):
                    in_node_index[tmp_in_edge[1]] = idx
                    in_node_timestamp.append(tmp_in_edge[2]['time'])

                for idx,tmp_out_edge in enumerate(G.edges(node, data=True)):
                    out_node_index[tmp_out_edge[1]] = idx
                    out_node_timestamp.append(tmp_out_edge[2]['time'])
            
            in_node_timestamp  = np.array(in_node_timestamp)
            out_node_timestamp = np.array(out_node_timestamp)
            
            if context == 'closeness':
                prob = np.expand_dims(in_node_timestamp, axis=-1) - np.expand_dims(out_node_timestamp, axis=0)
                prob = np.abs(prob) 
                
                # get the rank of each row, and use that to assign the weight for each edge
                prob = np.stack([rankdata(-x,'dense') for x in prob], axis=0)  
            
            elif context == 'unbiased':
                prob = np.expand_dims(in_node_timestamp, axis=-1) - np.expand_dims(out_node_timestamp, axis=0)
                prob[:] = 1 # the previous line is used for get the right shape of the array            
                
            else:
                print('Unsupported temporal context.')
                sys.exit()
                
            self.P_t[node] = prob
                
        print('>>> Completed!\n')